/*
 * balls2.cc
 * 
 * Molecular Dynamics simulation of rigid body interaction. 
 * 
 * Copyright (c) 2005 by Wolfgang Wieser ] wwieser (a) gmx <*> de [ 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#define _LIB_SALLOC_DONT_DEFINE_OPERATOR_NEW
#include <lib/sconfig.h>

#include <Qt/qapplication.h>
#include <Qt/qpainter.h>
#include <Qt/qbrush.h>
#include <Qt/qlayout.h>
#include <Qt/qlabel.h>
#include <Qt/qcdestyle.h>
//#include <Qt/qevent.h>

#include <QTX/xpainter.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <lib/numerics/3d/vector3.h>
#include <lib/numerics/3d/plane3.h>
#include <lib/tl/linkedlist.h>
#include <lib/htime.h>


namespace NUM
{
typedef Vector3<fp> fp3;
enum Coordinates { X=0, Y=1, Z=2 };
}

#include "grid_accel.h"


using namespace QTX;

namespace NUM
{

#define USE_GRID 1

//==============================================================================

// Integration time step: 
const fp const_dt=0.00001/2;  // 0.00001

// Gravity acceleration constant. 
const fp const_g=100;

// Initial speed of the bullet ball. 
const fp3 bullet_v0(6*1,0,-8);

// Size and density of the sand balls: 
const fp sand_ball_rad=0.005/3;                      // FINAL: 0.005/3
const fp sand_ball_density=3000;
// Bullet ball size and denstiy: 
const fp bullet_ball_rad=6*sand_ball_rad;
const fp bullet_ball_density=4000;

// Number of sand ball layers. 
const int z_layers=20;                               // FINAL: 20

// Stretch balls from tightest density this much. 
const fp stretch_fact=1.08*1.1;  // FIXME: This is needed to prevent explosion. 

// coeff_k: elastic modulus 
const fp coeff_k = 1e6;
// delta_0: normalisation constant for non-linearity coeff_nu 
//      of the potential and for the dimension. "Typical ball overlap."
const fp delta_0 = 0.1*sand_ball_rad;
// coeff_nu: non-linearity of the repulsive potential. 
const fp coeff_nu=1.5;  // "elastic spheres in 3d"

// coeff_gamma_n: phenomenological normal viscous dissipation coeff. 
const fp coeff_gamma_n=0.2*10;
// coeff_phi: non-linearity of the dissipative force in normal 
//     direction. 
const fp coeff_phi=1;   // 0 for "linear spring"  <==(***)

// coeff_gamma_t: phenomenological tangential viscous dissipation/damping 
// coeff. Be careful: System will explode for too large values...
const fp coeff_gamma_t=0.1*10;         // <==(***)

// coeff_mu >= 0 relates tangential force to normal force; 
//      Coulomb's law. 
const fp coeff_mu=0.4;// <<<

//==============================================================================

static inline fp3 RandomInSphere()
{
	// This does not have to be particularly fast here. 
	fp x,y,z;
	do {
		x=(rand()/fp(RAND_MAX))*2-1;
		y=(rand()/fp(RAND_MAX))*2-1;
		z=(rand()/fp(RAND_MAX))*2-1;
	} while(x*x+y*y+z*z>(fp)1.0);
	return(fp3(x,y,z));
}


struct Ball : LinkedListBase<Ball>
{
	fp rad;    // radius
	fp m,I;    // mass, inertia
	fp3 r;     // position
	fp3 v;     // velocity
	fp3 w;     // angular velocity
	fp3 f;     // force
	fp3 nxf;   // normal x force
	int texture_no;  // number of texture
	GridAccelerator<Ball,0>::PList *pl;
	bool fixed;
	
	inline void *operator new[](size_t s)
		{  return(malloc(s));  }
	inline void operator delete[](void *ptr)
		{  free(ptr);  }
	
	inline Ball() :  LinkedListBase<Ball>() {}
	inline ~Ball() {}
};


class Simulator
{
	public:
		typedef GridAccelerator<Ball,0>::PList PList;
		
	private:
		GridAccelerator<Ball,0> grid;
		int nballs;
		
		int nticks;
		int ndumps;
		fp time;
		int started;
		
		inline int _ComputeInteraction(Ball *bi,Ball *bj)
		{
			if(bi==bj)  return(0);
			
			fp3 n = bi->r - bj->r;
			fp dist=n.len2();
			
			if(dist>=sqr(bi->rad+bj->rad))  return(0);
			
			dist=sqrt(dist);
			n*=1.0/dist;  //n.normalize();  Normal unit vector. 
			
			// Amount of overlap; positive, normalized. 
			fp delta_rel=(bi->rad+bj->rad-dist)/delta_0;
	//fprintf(stderr,"%g ",delta_rel);
	Assert(delta_rel>=0);
	//fprintf(stderr,".");
			
			// vc: Contact velocity. 
			fp3 vc = bi->v-bj->v - cross(bi->rad*bi->w+bj->rad*bj->w,n);
			fp3 vc_n = dot(vc,n) * n;
			fp3 vc_t = vc - vc_n;
			// Normalize vc_t to get t. If vc_t is 0, make t=0 as well. 
			fp3 t;
			{
				fp vlen=vc_t.len2();
				if(vlen<1e-40)  t.promote(0);
				else  t=normalize(vc_t);
			}
			
			//fprintf(stderr,"%g %g %g\n",vc_t[0],vc_t[1],vc_t[2]);
			
			// Compute force: First, normal force which is: 
			//   1. repulsive force. 
			//   2. dissipation in normal direction (viscous damping force). 
			fp3 f_n = 
				( coeff_k*delta_0*pow(delta_rel,coeff_nu) 
				- coeff_gamma_n*dot(vc,n)*pow(delta_rel,coeff_phi) ) * n;
			// Compute tangential friction force: 
			//fp3 f_t = -coeff_gamma_t*vc_t;
			fp3 f_t = -min( coeff_gamma_t*dot(vc,t), coeff_mu*f_n.len() ) * t;
			
			// Total force: 
			fp3 f = f_n + f_t;
			
			// Sum up forces: 
			bi->f+=f;
			bi->nxf+=cross(n,f);
			
			// Double force if partner is fixed. 
			// FIXME: This only works for equally-sized balls (?!)
			if(bj->fixed)
			{
				bi->f+=f;
				bi->nxf+=cross(n,f);
			}
			
			return(1);
		}
		
		inline void _UpdateState(Ball *bi,fp dt)
		{
			bi->w -= (bi->rad/bi->I) * bi->nxf * dt;
			bi->v += bi->f/bi->m * dt;   // dP/m
			bi->r += bi->v * dt;
		}
		
	protected:
		XPainter *p_black;
		
		// Re-draw everything: 
		void _DrawAll();
		
	public:
		Simulator();
		~Simulator();
		
		void FillSand();
		
		void SimTick();
		
		void DumpState();
		
		void _IgniteBall();
		void _ResetVelocities();
};


void Simulator::FillSand()
{
	printf("sizeof(fp)=%d\n",sizeof(fp));
	const fp sqrt2=sqrt(2.0);
	const fp sqrt3=sqrt(3.0);
	
	//                  z
	// *            *    A
	// *            *    |
	// **************    +----> x
	
	fp3 universe_size(0.12,0.12,0.13);
	
	grid.SetCoSy(fp3(-universe_size[0]/2*0,-universe_size[1]/2*0,0),
		universe_size*stretch_fact,
		15,15,12 /*6,6,12*/);
	
	// Size and mass of the sand balls: 
	fp ball_rad=sand_ball_rad;
	fp ball_m=sand_ball_density*4.0/3.0*3.141592654*pow(sand_ball_rad,3.0);
	
	int nballs_x=(int)(universe_size[0]/(2*ball_rad));
	int nballs_y=(int)(universe_size[1]/(sqrt3*ball_rad));
	int nballs_z=z_layers;  // "Layers"
	printf("nballs= %d %d %d = %d\n",nballs_x,nballs_y,nballs_z,
		nballs_x*nballs_y*nballs_z);
	
	// Number of extra balls: 
	int nballs_extra=1;
	
	grid.AllocParticles(nballs_x*nballs_y*nballs_z+nballs_extra);
	nballs=0;
	
	// Extra ball: NOTE: It must be the first one!
	{
		Ball *b=&grid[nballs++];
		b->rad=bullet_ball_rad;
		b->m=bullet_ball_density*4.0/3.0*3.141592654*pow(bullet_ball_rad,3.0);
		b->I=0.4*b->m*sqr(b->rad);
		b->r=fp3(universe_size[0]/2*.4/*<<<*/  ,universe_size[1]/2,
			(1.1*b->rad+nballs_z*1.7*ball_rad));
		b->r*=(1+0.8*(stretch_fact-1));
		b->v=fp3(0,0,0); //bullet_v0;
		b->fixed=1;     // untils started
		b->texture_no=0;
		
		b->pl=grid.insert(b);
	}
	
	for(int z=0; z<nballs_z; z++)
	{
		for(int y=0; y<nballs_y; y++)
		{
			for(int x=0; x<nballs_x; x++)
			{
				int y_odd = (y+1000)%2;
				int z_odd = (z+1000)%2;
				
				Ball *b=&grid[nballs];
			b->fixed = (x==0 || x==nballs_x-1 || y==0 || y==nballs_y-1 || z==0);
			//b->fixed=0;
				//if(!b->fixed) continue;
				b->rad=ball_rad;
				b->m=ball_m;
				b->I=0.4*b->m*sqr(b->rad);
				b->r=fp3(
					(2*x + y_odd - z_odd) * ball_rad,
					(sqrt3*y + 0.5*sqrt3*z_odd) * ball_rad,
					(2.0*sqrt2/sqrt3*z) * ball_rad );
				//b->r -= fp3(universe_size[0]/2,universe_size[1]/2,0);
				//b->r[2] += ball_rad;
				b->r*=stretch_fact;
				// Displace balls slightly. 
				b->r+=RandomInSphere() * ball_rad * 0.05;
				b->v=fp3(0,0,0);
				b->texture_no=1;
				
				++nballs;
				b->pl=grid.insert(b);
			}
		}
	}
	
	printf("outside grid cell: %d balls\n",grid.list_outside()->count());
	//grid.DumpGridStat();
}


void Simulator::_IgniteBall()
{
	// Ignite ball. 
	Ball *b=&grid[0];
	
	b->v=bullet_v0;
	b->fixed=0;
	
	started=1;
}


void Simulator::_ResetVelocities()
{
	for(int i=0; i<nballs; i++)
	{
		Ball *bi=&grid[i];
		
		bi->v=fp3(0,0,0);
	}
}


void Simulator::_DrawAll()
{
	p_black->ClearWindow();
	
	int win_w=500;
	int win_h=500;
	int win_x0=50;
	int win_y0=win_h+50;
	
	fp3 off=grid.UOrigin();
	fp3 scale(win_w/grid.USize()[0],1,win_h/grid.USize()[2]);
	
	for(int i=0; i<nballs; i++)
	{
		Ball *b=&grid[i];
		
		int gx,gy,gz;
		grid.u2gl(b->r, gx,gy,gz);
		if(gy<grid.gsize_y()/2-1 || gy>grid.gsize_y()/2+1) continue;
		
		fp3 r=b->r-off;
		// Project x and z. 
		int xco = win_x0 + int( r[0]*scale[0] );
		int yco = win_y0 - int( r[2]*scale[2] );
		int width =  int(2*b->rad*scale[0]);
		int height = int(2*b->rad*scale[2]);
		
		p_black->DrawArc(xco-width/2,yco-height/2,width,height);
	}
}



void Simulator::SimTick()
{
	HTime start_time(HTime::Curr);
	
	fp dt=const_dt;
	
	// Reset force; compute gravity part. 
	for(int i=0; i<nballs; i++)
	{
		Ball *bi=&grid[i];
		if(bi->fixed) continue;
		
		bi->f.promote(0);
		bi->nxf.promote(0);
		
		bi->f[2]-=const_g*bi->m;
	}
	
	int ninteract=0;
	
	// Calculate the force. 
	for(int i=0; i<nballs; i++)
	{
		Ball *bi=&grid[i];
		if(bi->fixed) continue;
		
#if !USE_GRID
		// Dump O(N^2) algo. 
		for(int j=0; j<nballs; j++)
		{
			Ball *bj=&grid[j];
			
			ninteract+=_ComputeInteraction(bi,bj);
		}
#else
		// More intelligent algo. We assume that the first ball is the 
		// large bullet ball. 
		// For the bullet ball, do a normal (dump) step. This is O(N). 
		if(i==0)
		{
			for(int j=0; j<nballs; j++)
			{
				Ball *bj=&grid[j];
				
				ninteract+=_ComputeInteraction(bi,bj);
			}
			continue;
		}
		// For all other balls, compute interaction with bullet ball first: 
		_ComputeInteraction(bi,&grid[0]);
		
		// Then, compute interaction with all other balls. 
		// This is pretty fast since all the involved balls are quite small. 
		// Just make sure to exclude the bullet ball so that we do not 
		// count it twice. 
		
		// Determine which grid elements have to be checked at all. 
		// Therefore, compute grid coordinates of balls at locations 
		// bi->r - sand_ball_rad*x,y,z ... And, add some safety. 
		fp3 r_delta;  r_delta.promote(sand_ball_rad*2*1.1);
		int gx0,gy0,gz0;
		grid.u2gl(bi->r-r_delta, gx0,gy0,gz0);
		int gx1,gy1,gz1;
		grid.u2gl(bi->r+r_delta, gx1,gy1,gz1);
		
		bool outside=0;
		if(gx0<0)  {  gx0=0;  outside=1;  }
		if(gy0<0)  {  gy0=0;  outside=1;  }
		if(gz0<0)  {  gz0=0;  outside=1;  }
		if(gx0>=grid.gsize_x())  {  gx0=grid.gsize_x()-1;  outside=1;  }
		if(gy0>=grid.gsize_y())  {  gy0=grid.gsize_y()-1;  outside=1;  }
		if(gz0>=grid.gsize_z())  {  gz0=grid.gsize_z()-1;  outside=1;  }
		
		if(gx1<0)  {  gx1=0;  outside=1;  }
		if(gy1<0)  {  gy1=0;  outside=1;  }
		if(gz1<0)  {  gz1=0;  outside=1;  }
		if(gx1>=grid.gsize_x())  {  gx1=grid.gsize_x()-1;  outside=1;  }
		if(gy1>=grid.gsize_y())  {  gy1=grid.gsize_y()-1;  outside=1;  }
		if(gz1>=grid.gsize_z())  {  gz1=grid.gsize_z()-1;  outside=1;  }
		
		for(int gzi=gz0; gzi<=gz1; gzi++)
			for(int gyi=gy0; gyi<=gy1; gyi++)
				for(int gxi=gx0; gxi<=gx1; gxi++)
					for(Ball *bj=grid.lists_idx(gxi,gyi,gzi)->first(); 
						bj; bj=bj->next)
					{
						if(bj!=&grid[0]) // Bullet ball already handeled. 
						{
							ninteract+=_ComputeInteraction(bi,bj);
						}
					}
		if(outside)
			for(Ball *bj=grid.list_outside()->first(); bj; bj=bj->next)
			{
				if(bj!=&grid[0]) // Bullet ball already handeled. 
				{
					ninteract+=_ComputeInteraction(bi,bj);
				}
			}
#endif
	}
	
	// Update velocities and positions: 
	fp E_kin_tot=0;
	for(int i=0; i<nballs; i++)
	{
		Ball *bi=&grid[i];
		if(bi->fixed) continue;
		
		_UpdateState(bi,dt);
		
		// FIXME: Should add rotation energy. 
		//        bi->I*bi->w.len2()    (div by 2 lateron)
		E_kin_tot+=bi->v.len2()*bi->m;
		
		PList *pl_new=grid.u2gl(bi);
		if(pl_new!=bi->pl)
		{
			grid.ChangeGridCell(bi,bi->pl,pl_new);
			bi->pl=pl_new;
		}
	}
	
	E_kin_tot*=0.5;
	
	time+=dt;
	
	if(!(nticks++%20))
	{
		HTime end_time(HTime::Curr);
		HTime elapsed=end_time-start_time;
		fprintf(stderr,"coll=%d usec=%ld time=%g E_kin=%g n=%d    "
			"\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b",
			ninteract,elapsed.Get(HTime::usec),time,E_kin_tot,nticks-1);
		if(nticks==1)  fprintf(stderr,"\n");
		
		_DrawAll();
		
		if(started)
		{  DumpState();  }
	}
	
	// Auto-igition. 
	if(E_kin_tot<0.01 && nticks>5000 && ninteract>10000 && !started)
	{
		fprintf(stderr,"\n");
		_ResetVelocities();
		_IgniteBall();
	}
	
	if(ndumps>24*18)
	{
		fprintf(stderr,"\nExiting after %d dumps.\n",ndumps);
		exit(0);
	}
}


void Simulator::DumpState()
{
	char filename[256];
	snprintf(filename,256,"POV/f%07d.pov",ndumps++);
	
	FILE *fp=fopen(filename,"w");
	if(!fp)
	{  fprintf(stderr,"Opening %s: %s\n",filename,strerror(ferror(fp)));
		return;  }
	
	fprintf(fp,"// Frame %d, tick %d, time %g\n\n",ndumps-1,nticks,time);
	
	fprintf(fp,"#declare frame_no=%d;\n",ndumps-1);
	fprintf(fp,"#include \"../head.inc\"\n\n");
	
	fprintf(fp,"#macro MainBall()\n");
	{
		Ball *bi=&grid[0];
		fprintf(fp,"\tMBall(<%g,%g,%g>,<%g,%g,%g>,<%g,%g,%g>,<%g,%g,%g>,%g,%d,%d, %g)\n",
				bi->r[X],bi->r[Y],bi->r[Z],
				bi->v[X],bi->v[Y],bi->v[Z],
				bi->w[X],bi->w[Y],bi->w[Z],
				bi->f[X],bi->f[Y],bi->f[Z],
				bi->rad,
				bi->texture_no,
				bi->fixed ? 1 : 0,
				0.5 * (bi->v.len2()*bi->m + bi->w.len2()*bi->I));
	}
	fprintf(fp,"\n#end\n\n");
	
	fprintf(fp,"#macro BallList()\n");
	for(int i=1; i<nballs; i++)
	{
		Ball *bi=&grid[i];
		
		fprintf(fp,"\tBall(<%g,%g,%g>,<%g,%g,%g>,<%g,%g,%g>,<%g,%g,%g>,%g,%d,%d, %g)\n",
			bi->r[X],bi->r[Y],bi->r[Z],
			bi->v[X],bi->v[Y],bi->v[Z],
			bi->w[X],bi->w[Y],bi->w[Z],
			bi->f[X],bi->f[Y],bi->f[Z],
			bi->rad,
			bi->texture_no,
			bi->fixed ? 1 : 0,
			0.5 * (bi->v.len2()*bi->m + bi->w.len2()*bi->I));
	}
	fprintf(fp,"\n#end\n\n");
	
	fprintf(fp,"#include \"../main.inc\"\n");
	
	fclose(fp);
}


Simulator::Simulator() : 
	grid()
{
	p_black=NULL;
	nticks=0;
	ndumps=0;
	time=0;
	started=0;
	
	FillSand();
}

Simulator::~Simulator()
{
	if(p_black)
	{  p_black->~XPainter(); free(p_black);  }
}


}  // End of namespace NUM. 



class MainWindow : 
	public QWidget,
	public NUM::Simulator
{
	private:
		void timerEvent(QTimerEvent *);
		void keyPressEvent(QKeyEvent *);
	public:
		MainWindow(QWidget *parent=NULL,const char *name=NULL);
		~MainWindow();
};


void MainWindow::timerEvent(QTimerEvent *)
{
	SimTick();
}

void MainWindow::keyPressEvent(QKeyEvent *ev)
{
	switch(ev->key())
	{
		case Key_Return:
		case Key_Enter:  break;
		case Key_Q:  close();             break;
		case Key_D:  DumpState();         break;
		case Key_R:  _DrawAll();          break;
		case Key_S:  fprintf(stderr,"\n");  _IgniteBall();  break;
		case Key_V:  _ResetVelocities();  break;
		case Key_H:
			printf(
				"q -> close\n"
				"d -> dump state\n"
				"r -> redraw\n"
				"s -> start: ignite ball and start recording\n"
				"v -> reset all velocities (cool down)\n");
			break;
	}
}


MainWindow::MainWindow(QWidget *parent,const char *name) : 
	QWidget(parent,name)
{
	resize(600,600);
	move(400,100);
	
	setBackgroundColor(Qt::white);
	setCaption("Balls");
	
	QPainter pnt(this);
	pnt.setBrush(QBrush(Qt::black,Qt::SolidPattern));
	p_black=new(malloc(sizeof(XPainter))) XPainter(&pnt,true);
	Assert(p_black);
	
	startTimer(10);
	
	show();
}


MainWindow::~MainWindow()
{
}


int main(int argc,char **arg)
{
	QApplication qapp(argc,arg);
	qapp.setStyle(new(malloc(sizeof(QCDEStyle))) QCDEStyle(TRUE));
	
	MainWindow mainwin(NULL);
	qapp.setMainWidget(&mainwin);
	
	int r=qapp.exec();
	return(r);
}
